# Generated by Django 2.2.6 on 2020-12-16 02:56

from django.db import migrations


def reverse_func(apps, schema_editor):
    raise Exception("task migrate cannot reverse")


def forward_func(apps, schema_editor):
    from pipeline.contrib.periodic_task.djcelery import migrate

    migrate.try_to_migrate_to_django_celery_beat()


class Migration(migrations.Migration):

    dependencies = [
        ("periodic_task", "0004_auto_20191213_0828"),
    ]

    operations = [migrations.RunPython(forward_func, reverse_func)]
