# -*- coding: utf-8 -*-
# Generated by Django 1.11.29 on 2020-11-09 11:01
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline_external_plugins", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="s3source", name="source_dir", field=models.TextField(default="", verbose_name="源目录名"),
        ),
    ]
